# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


class ModuleDocFragment(object):
    DOCUMENTATION = r'''
options: {}
deprecated:
    alternative: Use some other module
    why: Test deprecation
    removed_in: '3.0.0'
'''
